;;; slack-authorize.el ---                           -*- lexical-binding: t; -*-

;; Copyright (C) 2018

;; Author:  <yuya373@archlinux>
;; Keywords:

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

;;; Commentary:

;;

;;; Code:
(require 'slack-util)
(require 'slack-team)
(require 'slack-request)
(require 'slack-channel)
(require 'slack-group)
(require 'slack-im)
(require 'slack-user)
(require 'slack-emoji)
(require 'slack-slash-commands)
(require 'slack-usergroup)
(require 'slack-message-notification)
(require 'slack-dnd-status)
(require 'slack-modeline)


(provide 'slack-authorize)
;;; slack-authorize.el ends here
